//----------------------------------------------------------------------------
// 
//	file: netlayer.cc
// 
//	function: network layer routines : implements a single neural 
//		network layer 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Wed Jul 22 04:37:01 CDT 1998
// 
//	$Revision: 1.1 $ $Date: 1998/09/04 03:36:37 $
//	$Log: netlayer-outstar.cc,v $
//	Revision 1.1  1998/09/04 03:36:37  yschoe
//	fixed all warnings generated by -Wall
//
//	Revision 1.12  1998/08/14 17:44:29  yschoe
//	1. fixed bug in resize_lat_exc_rad().
//	2. added kill_lat_weights.
//
//	Revision 1.11  1998/08/14 06:50:29  yschoe
//	1. fixed little cosmetic bug in destructor
//	2. added resize_lat_exc_rad()
//
//	Revision 1.10  1998/08/11 17:01:09  yschoe
//	1. Modularized norm_sum() and norm_factor() for easier
//	extension of weight normalization.
//	2. Added reset_matrix() and reset_output()
//
//	Revision 1.9  1998/08/02 11:05:30  yschoe
//	1. calc_netinput() separated from activate().
//	2. fixed set_parameter(), get_parameter() and added different interfaces.
//
//	Revision 1.8  1998/07/30 00:18:53  yschoe
//	1. fixed destructor, init_weights(),
//	   normalize_weights() to use precalculated ranges.
//	2. fixed serious bug in activate(), weighted sum loop.
//	3. added learn(), delta(), and get_parameter().
//
//	Revision 1.7  1998/07/28 06:44:22  yschoe
//	1. init_weights() can now take a range [low,high)
//	2. activate() was added
//	3. set_parameter() was added --> will be removed later (maybe)
//
//	Revision 1.6  1998/07/26 23:58:50  yschoe
//	1. fixed bug : initializing aff_..[].
//	2. added update_blackboard() for setting up afferent weight entries
//	   when new input layer was added.
//	3. fixed net_bb->add (see 1.)
//	4. added normalize_weights().
//	5. separated calc_rf_ranges from add_inp_layer() so that
//	   it could be done before memory allocation (in alloc_first()).
//
//	Revision 1.5  1998/07/26 02:11:55  yschoe
//	init_weights(), act_func() added.
//	sig_low, sig_high added and added to blackboard.
//	must add normalization code.
//
//	Revision 1.4  1998/07/24 11:22:08  yschoe
//	save() load() and list() blackboard
//	routines added
//
//	Revision 1.3  1998/07/24 08:22:09  yschoe
//	fixed bug in add_inp_layer.
//	added add_lat_connections.
//	variables were added to blackboard(this was made a separate function)
//
//	Revision 1.2  1998/07/23 08:45:32  yschoe
//	alloc/dealloc of weights, setting up variables done. 
//	Todo: adding everything to blackboard, initializing weights, 
//	and all the real action stuff.
//
//	Revision 1.1  1998/07/23 01:07:05  yschoe
//	jul22 stable version (except netlayer)
//
//
//----------------------------------------------------------------------------

#include "netlayer.h"
#include "parser.h"
#include "string.h"

//----------------------------------------------------------------------------
//	netlayer() 
//	  - Constructor
//	  - Only deal with stuff that is directly related to the 
//	  	network topology. Drop all stuff such as threshold, etc.
//----------------------------------------------------------------------------
netlayer::netlayer(char *name, int cols, int rows, 
		 int exc_rad=0, int inh_rad=0,
                 int max_inp_layers=0, netlayer *feeding = (netlayer *)NULL,
		 int inp_cols=0, int inp_rows=0, int  aff_rad=0) {

  strncpy(netlayer::name,name,strlen(name)+1);

  // network dimension
  netlayer::cols = cols; 
  netlayer::rows = rows; 

  // number of input layers and pointer to those areas.
  netlayer::max_inp_layers = max_inp_layers;
  netlayer::num_inp_layers = 0;

  // RF and lateral connection radius
  netlayer::lat_exc_rad = exc_rad;
  netlayer::lat_inh_rad = inh_rad;

  // check for input/output dimension restrictions
  
  if (inp_cols!=0 && inp_rows!=0) {
     if (MAX(inp_cols,cols)%MIN(inp_cols,cols)!=0 ||
         MAX(inp_rows,rows)%MIN(inp_rows,rows)!=0 ) {
	error_exit("netlayer::netlayer","input area and output area size "
		"must be of integral proportion.");
     }
  }

  // These are needed before anything is allocated 
  //	- taken out from alloc_first()
  netlayer::aff_rads 	= new int [max_inp_layers];
  netlayer::rf_ranges   = new NET_RF_Range** [max_inp_layers];

  // calculate afferent RF ranges first -- before allocating memory
  // 	- taken out from add_inp_layer()
  calc_rf_ranges(num_inp_layers,cols,rows,inp_cols,inp_rows,aff_rad);

  // allocate memory according to given network dimensions
  alloc_first(cols, rows, max_inp_layers, exc_rad, inh_rad); 

  // add connection to the given input layer : allocation is done within
  if (feeding != NULL) {
  	add_inp_layer(feeding,inp_cols,inp_rows,aff_rad);
  }

  // add all variables to the blackboard
  init_blackboard(name);

}

//----------------------------------------------------------------------------
//	~netlayer() 
//	  - descructor
//----------------------------------------------------------------------------
netlayer::~netlayer() {

  // dealloc stuff allocated by alloc_first()
  delete_matrix(cols,rows,net_input);
  delete_matrix(cols,rows,output);
  
  // dealloc afferent weights
  delete [] aff_rads;
  delete [] aff_contris;
  delete [] aff_noises;
  delete [] alpha_affs;
  delete [] feedings; // don't dealloc what's pointed by feedings[]
  for (int k=0; k<num_inp_layers; ++k) 
     for (int i=0; i<cols; ++i) 
	for (int j=0; j<rows; ++j) 
	   delete_matrix(rf_ranges[k][i][j].cols,
			 rf_ranges[k][i][j].rows,wgt[i][j].aff_ws[k]);

  // dealloc lateral weights
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j) {
	// MUST CHANGE THIS!
	delete_matrix(lat_exc_range[i][j].cols,
		      lat_exc_range[i][j].rows,wgt[i][j].lat_exc_w);
	delete_matrix(lat_inh_range[i][j].cols,
		      lat_inh_range[i][j].rows,wgt[i][j].lat_inh_w);
     }
  
  // finally, dealloc the master weight matrix
  delete_matrix(cols,rows,wgt);

  // delete RF range
  for (int i=0; i<max_inp_layers; ++i) delete_matrix(cols,rows,rf_ranges[i]);
  delete [] rf_ranges;

  // dealloc blackboard
  delete net_bb;  
} 

//----------------------------------------------------------------------------
//	alloc_matrix() 
//	  - allocate matrix (or sepcified type, or REAL) and return pointer
//----------------------------------------------------------------------------
REAL **netlayer::alloc_matrix(int cols, int rows) {

  REAL **rp;

  rp = new REAL* [cols];

  for (int i=0; i<cols; ++i) rp[i] = new REAL [rows];

  return rp;

}

//----------------------------------------------------------------------------
// Third argument is just used to specify the return type
NET_RF_Range **netlayer::alloc_matrix(int cols, int rows, 
		NET_RF_Range **rf_range) {

  NET_RF_Range **rfp;

  rfp = new NET_RF_Range* [cols];
  for (int i=0; i<cols; ++i) rfp[i] = new NET_RF_Range[rows];

  return rfp;

}

//----------------------------------------------------------------------------
// Third argument is just used to specify the return type
NET_Weights **netlayer::alloc_matrix(int cols, int rows, NET_Weights **wgts) {

  NET_Weights **wp;

  wp = new NET_Weights* [cols];
  for (int i=0; i<cols; ++i) wp[i] = new NET_Weights [rows];

  return wp;

}

//----------------------------------------------------------------------------
//	delete_matrix() 
//	  - delete matrix (or sepcified type, or REAL) and return pointer
//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, REAL **rp) {

  for (int i=0; i<cols; ++i) delete [] rp[i];
  delete [] rp; 

}

//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, NET_RF_Range **rfp) {

  for (int i=0; i<cols; ++i) delete [] rfp[i];
  delete [] rfp;

}

//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, NET_Weights **wgts) {

  for (int i=0; i<cols; ++i) delete [] wgts[i];
  delete [] wgts;

}

//----------------------------------------------------------------------------
//	alloc_first() 
//	  - allocate as much as possible from info provided by the 
//	    constructor.
//----------------------------------------------------------------------------
void netlayer::alloc_first(int cols, int rows,
                 int max_inp_layers,
                 int exc_rad, int inh_rad) {
 
  // Now, allocate all relevant memory
  // Individual memory allocation within structured components 
  // should also be allocated here.
  //
  //	1. for afferent input parameters, etc.
  //
  netlayer::aff_contris	= new REAL [max_inp_layers];
  netlayer::aff_noises 	= new REAL [max_inp_layers];
  netlayer::alpha_affs	= new REAL [max_inp_layers];
  netlayer::feedings 	= new netlayer* [max_inp_layers];
  //
  //	2. for the network (matrices)
  //
  netlayer::net_input   = alloc_matrix(cols,rows);
  netlayer::output	= alloc_matrix(cols,rows);
  netlayer::wgt		= alloc_matrix(cols,rows,netlayer::wgt);
  //
  // 	3. for the afferent weights (set up a table of poiters first)
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	// don't allocate any memory for afferent inputs yet. Do that in 
	// 	add_inp_layer().
	netlayer::wgt[i][j].aff_ws    = new REAL** [max_inp_layers];
     }
  }

  // add lateral connections : this needs to be done only once.
  add_lat_connections();	

}

//----------------------------------------------------------------------------
//	add_inp_layer() 
//	  - allocate necessary memory and set link to input layer
//----------------------------------------------------------------------------
void netlayer::add_inp_layer(netlayer *feeding, 
	int inp_cols, int inp_rows, int aff_rad) {

  int cur_inp = num_inp_layers;

  // if it is not the first input layer being added, calc_rf_ranges
  // has not been done yet, so do that first.
  if (cur_inp!=0) calc_rf_ranges(cur_inp,cols,rows,inp_cols,inp_rows,aff_rad);

  netlayer::feedings[cur_inp] = feeding;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// Now allocate memory for the afferent weights
  	wgt[i][j].aff_ws[cur_inp] = alloc_matrix (rf_ranges[cur_inp][i][j].cols,
		rf_ranges[cur_inp][i][j].rows);
     }
  } 

  if (cur_inp!=0) update_blackboard(cur_inp);

  // increment current input count (it becomes 1 after initial add)
  ++num_inp_layers;

}

//----------------------------------------------------------------------------
//	init_blackboard() 
//	  - alloc and fill in the blackboard
//----------------------------------------------------------------------------
void netlayer::init_blackboard(char *name) {

  //	        basic/ lat conn  & aff conn & rf_ranges 
  int max_vars = 100 + cols*rows*(2+2*max_inp_layers);
  cout << "netlayer: max bb vars = " << max_vars << endl;

  // initialize new blackboard
  net_bb = new blackboard (name,max_vars);

  // Add to blackboard : other variables involving iterations will be
  // added in the subroutines called above.
  //
  // - add in the same order as in netlayer.h to enforce consistency
  //   and avoid confusion 
  //
  net_bb->add("cols",&cols);
  net_bb->add("rows",&rows);
  net_bb->add("max_inp_layers",&max_inp_layers);
  net_bb->add("num_inp_layers",&num_inp_layers);

  net_bb->add("aff_rads",aff_rads,max_inp_layers);
  net_bb->add("lat_exc_rad",&lat_exc_rad);
  net_bb->add("lat_inh_rad",&lat_inh_rad);

  net_bb->add("sig_low",&sig_low);
  net_bb->add("sig_high",&sig_high);

  net_bb->add("threshold",&threshold);

  net_bb->add("aff_contris",aff_contris,max_inp_layers);
  net_bb->add("lat_exc_contri",&lat_exc_contri);
  net_bb->add("lat_inh_contri",&lat_inh_contri);

  net_bb->add("aff_noises",aff_noises,max_inp_layers);
  net_bb->add("lat_exc_noise",&lat_exc_noise);
  net_bb->add("lat_inh_noise",&lat_inh_noise);

  net_bb->add("intrinsic_noise",&intrinsic_noise);

  net_bb->add("alpha_affs",alpha_affs,max_inp_layers);
  net_bb->add("alpha_lat_exc",&alpha_lat_exc);
  net_bb->add("alpha_lat_inh",&alpha_lat_inh);

  net_bb->add("net_input",net_input,cols,rows);
  net_bb->add("output",output,cols,rows);
  
  // the RF range and lateral ranges are pre-calculated in 
  // init_weights(), so there's no need to save this redundant
  // information.

  // now the interesting part : connection weights.
  char tag[BB_NAME_LEN];
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1. the afferent weights
  	for (int k=0; k<num_inp_layers; ++k) {
		sprintf(tag,"wgts[%d][%d].aff_ws[%d]",i,j,k);
		net_bb->add(tag,wgt[i][j].aff_ws[k],
			rf_ranges[k][i][j].cols,rf_ranges[k][i][j].rows);
	}

	// 2. lateral excitatory weights
	sprintf(tag,"wgts[%d][%d].lat_exc_w",i,j);
	net_bb->add(tag,wgt[i][j].lat_exc_w,
			lat_exc_range[i][j].cols,lat_exc_range[i][j].rows);

	// 3. lateral inhibitory weights
	sprintf(tag,"wgts[%d][%d].lat_inh_w",i,j);
	net_bb->add(tag,wgt[i][j].lat_inh_w,
			lat_inh_range[i][j].cols,lat_inh_range[i][j].rows);
     }
  }

  // print warning about link to feedings
  // cout << "netlayer: you have to manually set netlayer::feedings[]\n";
  // cout << "          after load()ing the blackboard image.\n"; 

}


//----------------------------------------------------------------------------
//	add_lat_connections() 
//	  - allocate memory for lateral connections
//----------------------------------------------------------------------------
void netlayer::add_lat_connections() {

  lat_exc_range = alloc_matrix(cols,rows,lat_exc_range);
  lat_inh_range = alloc_matrix(cols,rows,lat_inh_range);

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1.1 calculate lateral excitatory ranges
 	lat_exc_range[i][j].lowx = MAX(0,i-lat_exc_rad);
 	lat_exc_range[i][j].lowy = MAX(0,j-lat_exc_rad);
 	lat_exc_range[i][j].highx= MIN(cols-1,i+lat_exc_rad);
 	lat_exc_range[i][j].highy= MIN(rows-1,j+lat_exc_rad);
	lat_exc_range[i][j].cols = lat_exc_range[i][j].highx 
		- lat_exc_range[i][j].lowx + 1;
	lat_exc_range[i][j].rows = lat_exc_range[i][j].highy
	      	- lat_exc_range[i][j].lowy + 1;
 
	// 1.2 Now allocate memory for the lat_exc weights
  	wgt[i][j].lat_exc_w = alloc_matrix(lat_exc_range[i][j].cols,
		lat_exc_range[i][j].rows);

	// 2.1 calculate lateral inhibitory ranges
 	lat_inh_range[i][j].lowx = MAX(0,i-lat_inh_rad);
 	lat_inh_range[i][j].lowy = MAX(0,j-lat_inh_rad);
 	lat_inh_range[i][j].highx= MIN(cols-1,i+lat_inh_rad);
 	lat_inh_range[i][j].highy= MIN(rows-1,j+lat_inh_rad);
	lat_inh_range[i][j].cols = lat_inh_range[i][j].highx 
		- lat_inh_range[i][j].lowx + 1;
	lat_inh_range[i][j].rows = lat_inh_range[i][j].highy
	      	- lat_inh_range[i][j].lowy + 1;
 
	// 2.2 Now allocate memory for the lat_inh weights
  	wgt[i][j].lat_inh_w = alloc_matrix(lat_inh_range[i][j].cols,
		lat_inh_range[i][j].rows);
     }
  } 

}


//----------------------------------------------------------------------------
//	save() 
//	  - save blackboards, using "prefix" 
//----------------------------------------------------------------------------
void netlayer::save(char *prefix) {

  char bbname[BB_NAME_LEN];

  sprintf(bbname,"%s-%s.bbimg",prefix,name);
  net_bb->save(bbname);

}

//----------------------------------------------------------------------------
//	load() 
//	  - load blackboards, using "filename"
//----------------------------------------------------------------------------
void netlayer::load(char *prefix) {

  char bbname[BB_NAME_LEN];

  sprintf(bbname,"%s-%s.bbimg",prefix,name);
  net_bb->load(bbname);

}


//----------------------------------------------------------------------------
//	list() 
//	  - list blackboard
//----------------------------------------------------------------------------
void netlayer::list() {

  net_bb->list();

}


//----------------------------------------------------------------------------
//	act_func() 		    /~~~
//	  - activation function :__/ 
//				   | | 
//			    sig_low  sig_high
//----------------------------------------------------------------------------
inline REAL netlayer::act_func(REAL x) {

  if (x <= sig_low)
    	return 0.0;
  else if (x >= sig_high)
    	return 1.0;
  else 
  	return (x-sig_low)/(sig_high-sig_low);

}

//----------------------------------------------------------------------------
//	init_weights() 
//	  - initialize lateral and afferent weights
//	  - sets initial weight values to [low,high) 
//----------------------------------------------------------------------------
void netlayer::init_weights(REAL low, REAL high) {

  // 1. first, the afferent weights 
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j)
  	for (int k=0; k<num_inp_layers; ++k) { // for all input layers
	
	   for(int x=0; x<rf_ranges[k][i][j].cols; ++x)
	      for(int y=0; y<rf_ranges[k][i][j].rows; ++y) {
	         wgt[i][j].aff_ws[k][x][y] = (REAL) drand48()*(high-low)+low;
	 	 // cout << "init_weights() : " << wgt[i][j].aff_ws[k][x][y] << endl;
		 //printf("aff[%d] [%d][%d]--[%d][%d]\n",k,i,j,x,y);
	      }

	}

  // 2. then, the lateral weights.
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j) {

	// 2.1 excitatory
	for(int x=0; x<lat_exc_range[i][j].cols; ++x)
	   for(int y=0; y<lat_exc_range[i][j].rows; ++y) {
	      wgt[i][j].lat_exc_w[x][y]= (REAL) drand48()*(high-low)+low;
		 //printf("exc [%d][%d]--[%d][%d]\n",i,j,x,y);
	   }

	// 2.2 inhibitory
	for(int x=0; x<lat_inh_range[i][j].cols; ++x)
	   for(int y=0; y<lat_inh_range[i][j].rows; ++y) {
	      wgt[i][j].lat_inh_w[x][y]= (REAL) drand48()*(high-low)+low;
		 //printf("inh [%d][%d]--[%d][%d]\n",i,j,x,y);
	   }

	// reset output
	output[i][j] = 3.3;
     }

}

//----------------------------------------------------------------------------
//	normalize_weights() 
//	  - normalize aff and lat weights with sum of weight
//----------------------------------------------------------------------------
void netlayer::normalize_weights() {

  int topx, topy;

  // 1. first, the afferent weights 
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j)
  	for (int k=0; k<num_inp_layers; ++k) { // for all input layers
	
	   REAL sum = 0.0;
	   for(int x=0; x<rf_ranges[k][i][j].cols; ++x)
	      for(int y=0; y<rf_ranges[k][i][j].rows; ++y)
		 sum += norm_sum(wgt[i][j].aff_ws[k][x][y]);

	   if (fabs(sum)<=MIN_REAL) sum = 1.0;

	   REAL norm = norm_factor(sum);
	   for(int x=0; x<rf_ranges[k][i][j].cols; ++x)
	      for(int y=0; y<rf_ranges[k][i][j].rows; ++y)
	         wgt[i][j].aff_ws[k][x][y] /= norm;

	}

  // 2. then, the lateral weights.
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j) {


 	// 2.1 excitatory weights
	REAL sum = 0.0;
	for(int x=0; x<lat_exc_range[i][j].cols; ++x)
	   for(int y=0; y<lat_exc_range[i][j].rows; ++y)
		sum += wgt[i][j].lat_exc_w[x][y];

	if (fabs(sum)<=MIN_REAL) sum = 1.0;

	for(int x=0; x<lat_exc_range[i][j].cols; ++x)
	   for(int y=0; y<lat_exc_range[i][j].rows; ++y)
		wgt[i][j].lat_exc_w[x][y] /= sum;

 	// 2.2 inhibitory weights

	sum = 0.0;
	for(int x=0; x<lat_inh_range[i][j].cols; ++x)
	   for(int y=0; y<lat_inh_range[i][j].rows; ++y)
	      sum += wgt[i][j].lat_inh_w[x][y];

	if (fabs(sum)<=MIN_REAL) sum = 1.0;

	for(int x=0; x<lat_inh_range[i][j].cols; ++x)
	   for(int y=0; y<lat_inh_range[i][j].rows; ++y)
	      wgt[i][j].lat_inh_w[x][y] /= sum;
     }

}


//----------------------------------------------------------------------------
//	calc_rf_ranges() 
//	  - calculate afferent RF ranges and lateral ranges
//----------------------------------------------------------------------------
void netlayer::calc_rf_ranges(int cur_inp, 
		int cols, int rows, 
		int inp_cols, int inp_rows,
 		int aff_rad){


  // cout << "calc_rf_ranges: " << cur_inp << endl;
  // Before going to the next step, alloc and calculate RF ranges
  //
  // 	1. allocate memory need RF ranges for [cols][rows]
  //
  if (inp_cols==0 || inp_rows==0) { // in case there's no input layer
	return;
  } else {
  	rf_ranges[cur_inp] = alloc_matrix(cols,rows,rf_ranges[cur_inp]);
  }
  //
  //	2. calculate the RF ranges for the respective [cols][rows]
  //		- must deal with different sized input vs. current layer
  //
  REAL mag_x, mag_y;

  mag_x = (REAL)cols/(REAL)inp_cols; 
  mag_y = (REAL)rows/(REAL)inp_rows;

  aff_rads[cur_inp] = aff_rad;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {


  	int cent_x = (int)(i/mag_x);
	int cent_y = (int)(j/mag_y);
   	// printf("cent(%d,%d) : mag(%f,%f) :",cent_x,cent_y,mag_x,mag_y);
 	rf_ranges[cur_inp][i][j].lowx = MAX(0,cent_x-aff_rad);
 	rf_ranges[cur_inp][i][j].lowy = MAX(0,cent_y-aff_rad);
 	rf_ranges[cur_inp][i][j].highx= MIN(inp_cols-1,cent_x+aff_rad);
 	rf_ranges[cur_inp][i][j].highy= MIN(inp_rows-1,cent_y+aff_rad);
	rf_ranges[cur_inp][i][j].cols = rf_ranges[cur_inp][i][j].highx 
		- rf_ranges[cur_inp][i][j].lowx + 1;
	rf_ranges[cur_inp][i][j].rows = rf_ranges[cur_inp][i][j].highy
	      - rf_ranges[cur_inp][i][j].lowy + 1;

     }
  }

}

//----------------------------------------------------------------------------
//	update_blackboard() 
//	  - update blackboard after additional input layer was added
//	    with add_inp_layer()
//----------------------------------------------------------------------------
void netlayer::update_blackboard(int cur_inp) {

  char tag[BB_NAME_LEN];

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 1. the afferent weights
        sprintf(tag,"wgts[%d][%d].aff_ws[%d]",i,j,cur_inp);
        net_bb->add(tag,wgt[i][j].aff_ws[cur_inp],
               rf_ranges[cur_inp][i][j].cols,rf_ranges[cur_inp][i][j].rows);

     }
  }

}


//----------------------------------------------------------------------------
//	activate() 
//	  - activate map : output[][] is the final activity
//----------------------------------------------------------------------------
void netlayer::activate() {

  //
  // 1. Calculate netinput
  //
  calc_netinput();

  //
  // 2. Calculate output (by applying the activation function)
  //

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	output[i][j] = act_func(net_input[i][j]);
     }
  } // 2. done

}


//----------------------------------------------------------------------------
//	calc_netinput() 
//	  - calculate netinput
//----------------------------------------------------------------------------
void netlayer::calc_netinput() {

  REAL	aff_netinp, exc_netinp, inh_netinp;
  int	basex, basey;

  //
  // 0. clear net_input[i][j]
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	net_input[i][j] = 0.0;
     }
  }

  //
  // 1. Calculate netinput
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

  // 1.1 For all afferent weights of unit [i][j]

        for (int k=0; k<num_inp_layers; ++k) {
           aff_netinp = 0.0; 
  	   basex = rf_ranges[k][i][j].lowx;
	   basey = rf_ranges[k][i][j].lowy;
	   for (int x=rf_ranges[k][i][j].lowx; 
		    x<=rf_ranges[k][i][j].highx; ++x)	
	      for (int y=rf_ranges[k][i][j].lowy; 
		    y<=rf_ranges[k][i][j].highy; ++y)	
	   		aff_netinp += feedings[k]->output[x][y] 
				   * wgt[i][j].aff_ws[k][x-basex][y-basey];

	   net_input[i][j] += aff_contris[k]*aff_netinp;
	}


	// 1.1 done
	
  // 1.2 For all excitatory lateral connections
  // 	later than: v1.12 use outstar calculation here.

	// do outstar only if this unit was active
  	basex = lat_exc_range[i][j].lowx;
	basey = lat_exc_range[i][j].lowy;
	/* outstar
	if (output[i][j]>0.0) {
        for (int x=lat_exc_range[i][j].lowx; 
		 x<=lat_exc_range[i][j].highx; ++x) 
           for (int y=lat_exc_range[i][j].lowy; 	
		    y<=lat_exc_range[i][j].highy; ++y)     
		net_input[x][y] += lat_exc_contri * output[i][j] 
				* wgt[i][j].lat_exc_w[x-basex][y-basey];
	}
	*/

	/* instar */
	exc_netinp = 0.0;
        for (int x=lat_exc_range[i][j].lowx; 
		 x<=lat_exc_range[i][j].highx; ++x) 
           for (int y=lat_exc_range[i][j].lowy; 	
		    y<=lat_exc_range[i][j].highy; ++y)     
		exc_netinp += output[x][y]
			   * wgt[i][j].lat_exc_w[x-basex][y-basey];	

	net_input[i][j] += lat_exc_contri * exc_netinp;
	// 1.2 done

  // 1.3 For all inhibitory lateral connections
  // 	later than: v1.12 use outstar calculation here.

	// do outstar only if this unit was active
  	basex = lat_inh_range[i][j].lowx;
	basey = lat_inh_range[i][j].lowy;
	/* outstar
	if (output[i][j]>0.0) {
        for (int x=lat_inh_range[i][j].lowx; 
		 x<=lat_inh_range[i][j].highx; ++x) 
           for (int y=lat_inh_range[i][j].lowy; 
	 	 y<=lat_inh_range[i][j].highy; ++y)     
		net_input[x][y] -= lat_inh_contri * output[i][j] 
				* wgt[i][j].lat_inh_w[x-basex][y-basey];
	}
	*/

	/* instar */
	inh_netinp = 0.0;
        for (int x=lat_inh_range[i][j].lowx; 
		 x<=lat_inh_range[i][j].highx; ++x) 
           for (int y=lat_inh_range[i][j].lowy; 
	 	 y<=lat_inh_range[i][j].highy; ++y)     
		inh_netinp += output[x][y] 
			   * wgt[i][j].lat_inh_w[x-basex][y-basey];

	net_input[i][j] -= lat_inh_contri * inh_netinp;
	// 1.3 done
     }

  } // 1. done : net_input[i][j] has the net input now.

}

//----------------------------------------------------------------------------
//	set_parameter() 
//	  - set parameter through a command language-like text interface
//	  - this is temporary
//	  - command syntax :  set <param_name>{[<col>]{[<row>]}} = <value>
//		'{ }' : optional
//----------------------------------------------------------------------------
void netlayer::set_parameter(char *cmd) {
   
  char command[BB_NAME_LEN];
  char param_name[BB_NAME_LEN]; 
  int  cols, rows, param_idx;
  int  int_val;
  REAL real_val;

  strncpy(command,cmd,strlen(cmd)+1);

  parser *cmdp = new parser(cmd);
  
  if (!cmdp->match("set")) return;

  cmdp->match_param(param_name,&cols,&rows);

  if (!cmdp->match("=")) return;

  param_idx = net_bb->find(param_name);

  if (param_idx < 0) {
	cout << "netlayer::set_parameter() : no such parameter " << param_name
	     << endl;
 	return;
  }

  switch (net_bb->table[param_idx].type) {

	case BB_INT: cmdp->scan_int(&int_val);
		*(net_bb->table[param_idx].data.i) =  int_val;
		break;
        case BB_IP: cmdp->scan_int(&int_val);
		net_bb->table[param_idx].data.ip[cols] = int_val;
		break;
        case BB_IPP: cmdp->scan_int(&int_val);
		net_bb->table[param_idx].data.ipp[cols][rows] = int_val;
		break;
	case BB_REAL: cmdp->scan_real(&real_val);
		*(net_bb->table[param_idx].data.r) =  real_val;
		break;
        case BB_RP: cmdp->scan_real(&real_val);
		net_bb->table[param_idx].data.rp[cols] = real_val;
		break;
        case BB_RPP: cmdp->scan_real(&real_val);
		net_bb->table[param_idx].data.rpp[cols][rows]=real_val;
		break;

   }

   delete cmdp;
}
//----------------------------------------------------------------------------
// just take the param name and value
void netlayer::set_parameter(char *param_name, REAL value) {

  parser *cmdp = new parser(param_name);

  cmdp->match_param(param_name,&cols,&rows);

  int param_idx = net_bb->find(param_name);

  if (param_idx < 0) {
	cout << "netlayer::set_parameter() : no such parameter " << param_name
	     << endl;
 	return;
  }

  switch (net_bb->table[param_idx].type) {
      case BB_INT: *(net_bb->table[param_idx].data.i) = (int)value; break;
      case BB_IP: net_bb->table[param_idx].data.ip[cols] = (int)value; break;
      case BB_IPP: net_bb->table[param_idx].data.ipp[cols][rows] = (int)value; 
		break;
      case BB_REAL: *(net_bb->table[param_idx].data.r) = value; break;
      case BB_RP: net_bb->table[param_idx].data.rp[cols] = value; break;
      case BB_RPP: net_bb->table[param_idx].data.rpp[cols][rows] = value; break;
  }

  delete cmdp;

}
//----------------------------------------------------------------------------
// take base param name and column as an argument 
void netlayer::set_parameter(char *param_name, int col, REAL val) {
	
  char cmd_str[BB_NAME_LEN];

  sprintf(cmd_str,"set %s[%d] = %f",param_name,col,val);
  
  set_parameter(cmd_str);

}
  
//----------------------------------------------------------------------------
// take base param name and column/row as an argument 
void netlayer::set_parameter(char *param_name, int col, int row, REAL val) {
	
  char cmd_str[BB_NAME_LEN];

  sprintf(cmd_str,"set %s[%d][%d] = %f",param_name,col,row,val);
  
  set_parameter(cmd_str);

}
  

//----------------------------------------------------------------------------
//	learn() 
//	  - using delta(), modify weights.
//----------------------------------------------------------------------------
void netlayer::learn() {

  int basex, basey, basei, basej;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

  // 1.1 For all afferent weights of unit [i][j]

        for (int k=0; k<num_inp_layers; ++k) {
	   basex = rf_ranges[k][i][j].lowx;
	   basey = rf_ranges[k][i][j].lowy;
	   for (int x=0; x<rf_ranges[k][i][j].cols; ++x)	
	      for (int y=0; y<rf_ranges[k][i][j].rows; ++y)	
	         wgt[i][j].aff_ws[k][x][y] += delta(i,j,k,x+basex,y+basey);
	}
	
	// 1.1 done
	
  // 1.2 For all excitatory lateral connections

	//if (output[i][j]>0.0) {

	basex = lat_exc_range[i][j].lowx;
	basey = lat_exc_range[i][j].lowy;
        for (int x=0; x<lat_exc_range[i][j].cols; ++x) 
           for (int y=0; y<lat_exc_range[i][j].rows; ++y)
	      wgt[i][j].lat_exc_w[x][y] += delta(i,j,x+basex,y+basey,
						 alpha_lat_exc);

	//}
        // 1.2 done

  // 1.3 For all inhibitory lateral connections

	//if (output[i][j]>0.0) {

	basex = lat_inh_range[i][j].lowx;
	basey = lat_inh_range[i][j].lowy;
        for (int x=0; x<lat_inh_range[i][j].cols; ++x) 
           for (int y=0; y<lat_inh_range[i][j].rows; ++y)     
	      wgt[i][j].lat_inh_w[x][y] += delta(i,j,x+basex,y+basey,
						alpha_lat_inh);
	//}
	// 1.3 done
     }

  } // 1. done : all weights were modified -- it has yet to be normalized.

}


//----------------------------------------------------------------------------
//	delta() 
//	  - calculate delta, on which weight adaptation is based on
//----------------------------------------------------------------------------
inline REAL netlayer::delta(int col, int row, int cur, int x, int y) {

// for afferent weights : the current input layer is specified by 'cur_inp'

  return (alpha_affs[cur] * (output[col][row] * feedings[cur]->output[x][y]));

}
//----------------------------------------------------------------------------
inline REAL netlayer::delta(int col, int row, int x, int y, REAL alpha) {

// for lateral weights

  return (alpha * (output[col][row]*output[x][y]));

}

//----------------------------------------------------------------------------
//	get_parameter() 
//	  - get parameter value : always return in REAL. Only supports
//	    scalar variables. Returns 0.0 in all other cases.
//----------------------------------------------------------------------------
REAL netlayer::get_parameter(char *param) {

  char param_name[BB_NAME_LEN]; 
  int  cols, rows;
  int param_idx;

  parser *cmdp = new parser(param);

  cmdp->match_param(param_name,&cols,&rows);

  param_idx = net_bb->find(param_name);

  switch (net_bb->table[param_idx].type) {
      case BB_INT: return (REAL)(*(net_bb->table[param_idx].data.i));
      case BB_IP: return (REAL)(net_bb->table[param_idx].data.ip[cols]);
      case BB_IPP: return (REAL)(net_bb->table[param_idx].data.ipp[cols][rows]);
      case BB_REAL: return *(net_bb->table[param_idx].data.r);
      case BB_RP: return net_bb->table[param_idx].data.rp[cols];
      case BB_RPP: return net_bb->table[param_idx].data.rpp[cols][rows];
  }

  delete cmdp;

}
//----------------------------------------------------------------------------
REAL netlayer::get_parameter(char *param, int col) {
	
  char param_name[BB_NAME_LEN];

  sprintf(param_name,"%s[%d]",param,col);
  
  return get_parameter(param_name);

}
//----------------------------------------------------------------------------
REAL netlayer::get_parameter(char *param, int col, int row) {
	
  char param_name[BB_NAME_LEN];

  sprintf(param_name,"%s[%d][%d]",param,col,row);
  
  return get_parameter(param_name);

}


//----------------------------------------------------------------------------
//	reset_matrix() 
//	  - reset matrix to all zeros
//----------------------------------------------------------------------------
void netlayer::reset_matrix(int **mat, int cols, int rows) {

  for (int x=0; x<cols; ++x)
    for (int y=0; y<cols; ++y)
	mat[x][y] = 0;
    
}

//----------------------------------------------------------------------------
void netlayer::reset_matrix(REAL **mat, int cols, int rows) {

  for (int x=0; x<cols; ++x)
    for (int y=0; y<cols; ++y)
	mat[x][y] = 0.0;

}


//----------------------------------------------------------------------------
//	reset_output() 
//	  - reset output activity
//----------------------------------------------------------------------------
void netlayer::reset_output() {

   reset_matrix(output,cols,rows);

}


//----------------------------------------------------------------------------
//	resize_lat() 
//	  - resize lateral exc/inh range
//----------------------------------------------------------------------------
void netlayer::resize_lat_exc_rad(int new_rad) {

  // 1. calculate range and allocate new, while copying over
  //	old weights

  NET_RF_Range **lat_exc_range = alloc_matrix(cols,rows,lat_exc_range);
  REAL **temp_exc_w;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1.1 calculate lateral excitatory ranges
 	lat_exc_range[i][j].lowx = MAX(0,i-new_rad);
 	lat_exc_range[i][j].lowy = MAX(0,j-new_rad);
 	lat_exc_range[i][j].highx= MIN(cols-1,i+new_rad);
 	lat_exc_range[i][j].highy= MIN(rows-1,j+new_rad);
	lat_exc_range[i][j].cols = lat_exc_range[i][j].highx 
		- lat_exc_range[i][j].lowx + 1;
	lat_exc_range[i][j].rows = lat_exc_range[i][j].highy
	      	- lat_exc_range[i][j].lowy + 1;
 
	// 1.2 Now allocate memory for the lat_exc weights (in a temp
	//	variable temp_exc_w.
	temp_exc_w = alloc_matrix(lat_exc_range[i][j].cols,
                lat_exc_range[i][j].rows);

    	// 1.3 Clear new one's lat_exc weights
	for (int x=0; x<lat_exc_range[i][j].cols; ++x) 
	   for (int y=0; y<lat_exc_range[i][j].rows; ++y)
		temp_exc_w[x][y] = 0.0;
		
 	// 1.4 Now copy over the old weights to the new wgt matrix
 	int new_offset, old_offset;
 	if (lat_exc_rad > new_rad) {
	  old_offset = lat_exc_rad-new_rad; new_offset = 0;
	} else { 
	  old_offset = 0; new_offset = new_rad-lat_exc_rad;
	}
	for (int x=0; x<lat_exc_range[i][j].cols; ++x) 
	   for (int y=0; y<lat_exc_range[i][j].rows; ++y) 
	   	temp_exc_w[x+new_offset][y+new_offset] 
			= wgt[i][j].lat_exc_w[x+old_offset][y+old_offset];

	// 1.5 free the old weights and link it to the new one
	delete_matrix(netlayer::lat_exc_range[i][j].cols,
		netlayer::lat_exc_range[i][j].rows,wgt[i][j].lat_exc_w);
	wgt[i][j].lat_exc_w = temp_exc_w;

	// 1.7 update the blackboard (no need to free anything here) 
	char tag[BB_NAME_LEN];
	sprintf(tag,"wgts[%d][%d].lat_exc_w",i,j);
	net_bb->update(tag,temp_exc_w,
			lat_exc_range[i][j].cols, lat_exc_range[i][j].rows);
		
     }
  } 

  // 2. Free the old ranges and link it up to the new one
  delete_matrix(cols,rows,netlayer::lat_exc_range);
  netlayer::lat_exc_range = lat_exc_range;

  // 3. update lat_exc_rad
  netlayer::lat_exc_rad = new_rad;

}

//----------------------------------------------------------------------------
//	kill_lat_weights() 
//	  - kill weak connections
//----------------------------------------------------------------------------
void  netlayer::kill_lat_weights(REAL threshold) {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	// 1. for all exc weights -- skip this for now
	// 2. for all inh weights
	for (int x=0; x<lat_inh_range[i][j].cols; ++x) {
	   for (int y=0; y<lat_inh_range[i][j].rows; ++y) {
		if (wgt[i][j].lat_inh_w[x][y] <= threshold) {
		   wgt[i][j].lat_inh_w[x][y] = MIN_REAL;
		}
	   }
	}
	// 2. End
     }
  }

}
