//----------------------------------------------------------------------------
// 
//	file: plot_window.cc
// 
//	function: plot graph of datastream : x>=0, y>=0
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Thu Jul  9 05:01:22 CDT 1998
// 
//	$Revision: 1.4 $ $Date: 2000/02/13 10:29:43 $
//	$Log: plot_window.cc,v $
//	Revision 1.4  2000/02/13 10:29:43  yschoe
//	*** empty log message ***
//
//	Revision 1.3  1998/09/04 03:36:37  yschoe
//	fixed all warnings generated by -Wall
//
//	Revision 1.2  1998/08/11 17:07:10  yschoe
//	1.fixed min/max stuff for more generic use.
//	2. fixed nasty bug in redraw() that prevented some later instances
//	of plot_window not being redrawn.
//
//	Revision 1.1  1998/07/23 01:07:05  yschoe
//	jul22 stable version (except netlayer)
//
//
//----------------------------------------------------------------------------
#include <window.h>
#include <string.h>
#include <iostream.h>
#include <stream.h>
#include "plot_window.h"
#include "ppmout.h"

//----------------------------------------------------------------------------
//	plot_window() 
//	  - Constructors
//	  - max_points : max number of datapoints to hold
//	  - window_size : number of datapoints to display at a time
//	  - scale -- is used in a different context (from matrix_window)
//		  -- it multiplies window_size to change how many data points
//		     are displayed
//----------------------------------------------------------------------------
plot_window::plot_window
	(window &p, int w, int h, int x, int y, int bw, 
	int color_range, int color_mode, int max_points, int window_size)
: matrix_window(p,w,h,x,y,bw,color_range,color_mode,max_points,1) {

  plot_mode = ITER; // data is feeded one by one
  plot_window::window_size = window_size; // display window size

  plot_window::color_mode  = (ColMode)COLOR;
  plot_window::color_range = color_range; // number of colors to use
  plot_window::scale       = 1.0; // default scale 


  init(max_points);

  // Actually allocate colors
  alloc_colors();

}

//----------------------------------------------------------------------------
plot_window::plot_window
	(window &p, int w, int h, int x, int y, int bw, 
	int color_range, int color_mode, int max_points, REAL *oned)
: matrix_window(p,w,h,x,y,bw,color_range,color_mode,max_points,1) {

  plot_mode = BATCH; // a whole vector is presented as input
  plot_window::oned = oned; 
  window_size = max_points;

  plot_window::color_mode  = (ColMode)COLOR;
  plot_window::color_range = color_range; // number of colors to use
  plot_window::scale       = 1.0; // default scale 
 
  init(max_points);

  // Actually allocate colors
  alloc_colors();

}

//----------------------------------------------------------------------------
//	~plot_window() 
//	  - Destructor
//----------------------------------------------------------------------------
plot_window::~plot_window(){

  // XFreePixmap(display,pix); done elsewhere?
  delete [] color_map;
  // delete [] oned; DO NOT DELETE IT!

}

//----------------------------------------------------------------------------
//	init() 
//	  - init coordinate values etc. according to window width and height
//----------------------------------------------------------------------------
void plot_window::init(int max_points) {

  cerr << "matrix_window::init color_range = " << color_range << endl;

  col     = max_points;  // total datapoints to buffer
  row 	  = 1;		 // 1-D vector..

  cur_loc = 0;

  x_origin = 20;	// (x,y) origin for axis plotting
  y_origin = height-20; //
  x_size = width - 40;
  y_size = height - 40;

  x_scale = 1.0; // default x-axis scale
  y_scale = 1.0; // default y-axis scale
  zoom    = 1.0; // default zoom factor for window_size

  color_map = new int[color_range]; // allocate table for cmap
  plot_window::max_points = max_points;

  if (plot_mode == ITER)
  	oned = new REAL [max_points];

  y_max   = 0.00000001; // default y max

  if (plot_mode==BATCH) 	
	x_max   = (REAL) max_points; // default x max
  else 
	x_max   = (REAL) (window_size); // default x max

  // allocate other colors
  blue  = alloc_named_color("blue"); red   = alloc_named_color("red");
  green = alloc_named_color("green"); cyan  = alloc_named_color("cyan");
  
}

//----------------------------------------------------------------------------
//	DrawLine() 
//	  - draw line from src to dest : values are in cartesian coord
//----------------------------------------------------------------------------
void plot_window::DrawLine (REAL src_x, REAL src_y, REAL dest_x, REAL dest_y) {

  int x1, y1, x2, y2;
  REAL xgran,ygran;

  // set display granularity
  xgran = (REAL)x_size/x_max;
  ygran = (REAL)y_size/y_max;

  // set coordinates
  x1 = (int)(src_x*xgran)+x_origin;
  y1 = y_origin-(int)(src_y*ygran);
  x2 = (int)(dest_x*xgran)+x_origin;
  y2 = y_origin-(int)(dest_y*ygran);

  // Draw the line
  line(x1,y1,x2,y2);

}


//----------------------------------------------------------------------------
//	CopyArea() 
//	  - copy area -- take coordinates as arguments
//----------------------------------------------------------------------------
void plot_window::CopyArea(REAL sx,REAL sy,REAL ex,REAL ey,REAL dx,REAL dy){

  int x1, y1, x2, y2, dest_x, dest_y;
  REAL xgran,ygran;

  // set display granularity
  xgran = (REAL)x_size/x_max;
  ygran = (REAL)y_size/y_max;

  // set coordinates
  x1 = (int)(sx*xgran)+x_origin;
  y1 = y_origin-(int)(sy*ygran);
  x2 = (int)(ex*xgran)+x_origin;
  y2 = y_origin-(int)(ey*ygran);
  dest_x = (int)(dx*xgran)+x_origin;
  dest_y = y_origin-(int)(dy*ygran);

  XCopyArea(display,Win,Win,gc_copy,x1,y1,x2,y2,dest_x,dest_y);

}

//----------------------------------------------------------------------------
//	draw_interior() 
//	  - main drawing routine : use redraw() when you need to invoke this
//----------------------------------------------------------------------------
void plot_window::draw_interior() {

  int last_loc=0;
  char	str_buf[100];
  int zoom_window_size;

  //// find actual sliding window_size 
  // zoom_window_size = (int)(zoom*window_size);
  // x_max = zoom_window_size;
  zoom_window_size = window_size;
  //col = x_max;

  // draw x-axis
  set_color(blue);
  DrawLine(0.0,0.0,(REAL)x_max,0.0);

  // Find max y (for BATCH mode only) -- this is updated automatically in 
  // incremental mode.
  if (plot_mode==BATCH)  {
  	REAL max_val = -999999999.9;
	last_loc = max_points;
  	for (int i=cur_loc; i<last_loc; ++i) max_val = MAX(max_val,oned[i]);
  	y_max = max_val;
  } 

  // draw y-axis
  set_color(blue);
  DrawLine(0.0,0.0,0.0,(REAL)y_max);

  // plot all datapoints
  set_color(black);
  if (plot_mode==BATCH) {

  	for (int i=cur_loc; i<last_loc-1; ++i)
		DrawLine((REAL)i,oned[i],(REAL)i+1,oned[i+1]);

  } else if (plot_mode==ITER) {

	int left  = ((cur_loc-1)<zoom_window_size) ? 
			0 : ((cur_loc-1)%max_points)-zoom_window_size+1;
  	int right = ((cur_loc-1)%max_points);
  	int offset = 0;

//	if (cur_loc>=zoom_window_size) { // make it faster 
//		offset = zoom_window_size-1;
//		left = right-1;
//		CopyArea(1.0,0.0,zoom_window_size-1,y_max,0.0,0.0);
//		XCopyArea(display,pix,Win,gc_copy,20,0,width-20,height,0,0);
//	}

	for (int i=left; i<right; ++i) { // absolute location
		// printf("i=%d: (%d,%lf)\n",i,(i-left),oned[i%max_points]);
		DrawLine((REAL)(i-left+offset),oned[(i+max_points)%max_points],
		    (REAL)(i+1-left+offset),oned[(i+1+max_points)%max_points]);
		DrawLine((REAL)(i-left+offset),oned[(i+max_points)%max_points],
		    (REAL)(i-left+offset),0.0);
	}

  }

  // print max y value
  set_color(blue); sprintf(str_buf,"y=%.4f",y_max);
  PlaceText(str_buf,10,10);
  set_color(cyan); DrawLine(0.0,y_max*(1.0+1.0/y_size),
			x_max,y_max*(1.0+1.0/y_size)); 
  
  // print median x value
  set_color(blue); 
  REAL middle = (cur_loc>=zoom_window_size)
		? cur_loc-(zoom_window_size/2) : zoom_window_size/2;
  sprintf(str_buf,"%.0f",middle);
  PlaceText(str_buf,x_size/2+x_origin-x_size/30,y_origin+15);
  DrawLine(x_max/2.0,y_max/80.0,x_max/2.0,-y_max/80.0); 
  set_color(black); 

}

//----------------------------------------------------------------------------
//	set_scale() 
//	  - change the scale for x and y axis
//----------------------------------------------------------------------------
void plot_window::set_scale(REAL x, REAL y) {
  
  x_scale = x;
  y_scale = y;
  redraw();

}

//----------------------------------------------------------------------------
//	set_zoom() 
//	  - change the zoom factor for window_size
//----------------------------------------------------------------------------
void plot_window::set_zoom(REAL zoom) {
  
  plot_window::zoom = zoom;
  redraw();

}

//----------------------------------------------------------------------------
//	get_zoom() 
//	  - change the zoom factor for window_size
//----------------------------------------------------------------------------
REAL plot_window::get_zoom() {
  
  return zoom;

}

//----------------------------------------------------------------------------
//	disp_values() 	: let's not use this
//	  - dump matrix values to a text array
//----------------------------------------------------------------------------
char **plot_window::disp_values() {

  char **dump = new char* [row];
  char *buf  = new char [50];

  int cells=col*row;

  for (int i=0; i<row; ++i) {
	dump[i] = new char [col*10+1];	// 10 chars per value, and \n at end
  	strncpy(dump[i],"",0); // clear string
  }

  for (int i=0; i<cells; ++i)
  {
	if (dim_mode == OneD) {
		sprintf(buf,"%.4f ",oned[i]);
		strncat(dump[i/col],buf,strlen(buf));
		printf("%s",buf);
		if (i%col==0) printf("\n");
	} else if (dim_mode == TwoD) {
		sprintf(buf,"%.4f ",twod[i%col][i/col]);
		strncat(dump[i/col],buf,strlen(buf));
		printf("%s",buf);
		if (i%col==0) printf("\n");
	}
  }

  // delete [] dump;
  return dump;
}


//----------------------------------------------------------------------------
//	resize() 
//	  - resize and update some internal values as well
//----------------------------------------------------------------------------
void plot_window::resize(int width, int height) {

  matrix_window::resize(width,height);
  x_origin = 20;	// (x,y) origin for axis plotting
  y_origin = height-20; //
  x_size = width - 40;
  y_size = height - 40;
  redraw();

}

//----------------------------------------------------------------------------
//	push_data() 
//	  - push new data into the array
//----------------------------------------------------------------------------
void plot_window::push_data(REAL value) {

  oned[cur_loc%max_points] = value;
  y_max = MAX(y_max,value);
  cur_loc++;
  if (cur_loc%3==0) redraw();

}

//----------------------------------------------------------------------------
//	push_data() 
//	  - push new data into the array
//----------------------------------------------------------------------------
void plot_window::push_data(int value) {
  
  static int redraw_rate;

  oned[cur_loc%max_points] = (REAL) value;
  y_max = MAX(y_max,value);
  cur_loc++;
  if (redraw_rate++%3==0) redraw();

}

