## usage: [E,V,Emax,Vmax,Eimg] = convedgematnew1(Img, kernelsize, gridsize, freq)
##
##	Img: p x q matrix (grayscale image)
##
##		Img: 2D matrix with grayscale image
##		kernelsize: width (==height) of convolution kernel
##		gridsize:   sampling grid size
##		freq: 	    gabor frequency (relative to the kernelsize) 
##
## Returns a n x m matrix E where 
##	n = p / gridsize
##	m = q / gridsize
##	and each E_ij is the edge orientation of the subimage patch.
##
## First, convolve with DoG filter, and then convolve with gabor
## filters of orientations 0, 60, 120. Do a polar addition. Cut the generated
## images into patches and find the mximum value and the index	
##
##	Usage example:
##		[E,V,Emax,Vmax,Eimg] = convedgematnew1(imagemat,9,3,1);
##
##		both DoG and gabor filters are  9 x 9, and
##		Gabor frequency is 1.
##		final grid sampling size is 3 x 3.
##
## $Id: convedgemat.m,v 1.2 2002/03/09 17:31:27 yschoe Exp yschoe $

function [E,V,Emax,Vmax,Eimg] = convedgematnew1(Img, kernelsize, gridsize, freq)

	# get image size
	[oldp, oldq] = size(Img);

	# adjust grid size to have the conv kernels
	# centered at the middle: i.e. make the gridsize an odd number.
	if ( rem(kernelsize,2) == 0 ) 
		kernelsize ++;
	endif

        d = dogfilter(Img,kernelsize);
        
        
        # calculate new image size to make it divisible by gridsize
	p = floor(oldp/gridsize)*gridsize;
	q = floor(oldq/gridsize)*gridsize;

   	ygrids = floor(p/gridsize);
	xgrids = floor(q/gridsize);

	#initialize matrices
	V = zeros(ygrids,xgrids).-999999;
	E = zeros(ygrids,xgrids);
	
        Vmax=zeros(oldp,oldq);
	V1max=zeros(oldp,oldq);
        V2max=zeros(oldp,oldq);
        V3max=zeros(oldp,oldq);
        V4max=zeros(oldp,oldq);
        V5max=zeros(oldp,oldq);
     

        Emax=zeros(oldp,oldq);
       
        c1 = conv2(d,gabor(kernelsize,kernelsize,"cos",0,freq),"same");
	s1 = conv2(d,gabor(kernelsize,kernelsize,"sin",0,freq),"same");
	
        V1max= (real(c1)).^2 + (real(s1)).^2;

         c2 = conv2(d,gabor(kernelsize,kernelsize,"cos",36,freq),"same");
	s2 = conv2(d,gabor(kernelsize,kernelsize,"sin",36,freq),"same");
	
        V2max= (real(c2)).^2 + (real(s2)).^2;

         c3 = conv2(d,gabor(kernelsize,kernelsize,"cos",72,freq),"same");
	s3 = conv2(d,gabor(kernelsize,kernelsize,"sin",72,freq),"same");
	
        V3max= (real(c3)).^2 + (real(s3)).^2;

         c4 = conv2(d,gabor(kernelsize,kernelsize,"cos",108,freq),"same");
	s4 = conv2(d,gabor(kernelsize,kernelsize,"sin",108,freq),"same");
	
        V4max= (real(c4)).^2 + (real(s4)).^2;

         c5 = conv2(d,gabor(kernelsize,kernelsize,"cos",144,freq),"same");
	s5 = conv2(d,gabor(kernelsize,kernelsize,"sin",144,freq),"same");
	
        V5max= (real(c5)).^2 + (real(s5)).^2;

        v=[0,0];
        for i=1:oldp
	for j=1:oldq

        [p1,q1]=pol2cart(2.0*0*pi/180.0,V1max(i,j));
        [p2,q2]=pol2cart(2.0*36*pi/180.0,V2max(i,j));
        [p3,q3]=pol2cart(2.0*72*pi/180.0,V3max(i,j));
        [p4,q4]=pol2cart(2.0*108*pi/180.0,V4max(i,j));
        [p5,q5]=pol2cart(2.0*144*pi/180.0,V5max(i,j));
         
        v =  [p1,q1] + [p2,q2] + [p3,q3] + [p4,q4] + [p5,q5];
 
        [or,sel] = cart2pol(v(1),v(2));
        or = or*180.0/pi/2.0;
	Vmax(i,j)=sel;
        if (or<0)
	or = or+180;
        endif
        Emax(i,j) = round(or/10)+1;
     
        
        endfor
        endfor    

        #cut Vmax and Emax into grids and calculate V and E
 	for j=1:ygrids
          for i=1:xgrids
		vtmp = vec(Vmax(1+((j-1)*gridsize):(j*gridsize),
			        1+((i-1)*gridsize):(i*gridsize)));
		# idx contains the index of the max within that subimage
		# block.
  	        [V(j,i),idx]=max(vtmp);
		etmp = vec(Emax(1+((j-1)*gridsize):(j*gridsize),
				1+((i-1)*gridsize):(i*gridsize)));
                E(j,i)=etmp(idx);
          endfor
        endfor

endfunction
